# Imports from us-elections.
from elections.models.elections import PrimaryElection


# Imports from other dependencies.
import us


class PrimaryRunoffElection(PrimaryElection):
    is_runoff = True
    repr_label = "PrimaryRunoffElection"


class DemocraticPrimaryRunoffElection(PrimaryRunoffElection):
    repr_label = "DemocraticPrimaryRunoffElection"


class RepublicanPrimaryRunoffElection(PrimaryRunoffElection):
    repr_label = "RepublicanPrimaryRunoffElection"



# Imports from python.
from pkg_resources import resource_stream


# Imports from us-elections.
from elections.constants import DATA_MODULE_PATH


def load_pickled_data(pickled_file):
    return resource_stream(DATA_MODULE_PATH, pickled_file)



# Imports from other dependencies.
import us


supt_verbose = "Superintendent of Public Instruction"

EXECUTIVE_OFFICE_NAMES = {
    "president": "President",
    "vice-president": "Vice President",
    "governor": "Governor",
    "lieutenant-governor": "Lieutenant Governor",
    "secretary-of-state": "Secretary of State",
    "superintendent-of-public-instruction": supt_verbose,
    "attorney-general": "Attorney General",
    "auditor": "Auditor",
    "treasurer": "Treasurer",
    "agriculture-commissioner": "Commissioner of Agriculture",
    "labor-commissioner": "Commissioner of Labor",
    "insurance-commissioner": "Insurance Commissioner",
    "commissioner-of-public-lands": "Commissioner of Public Lands",
}


class ExecutiveSeat(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if k == "state":
                self.__dict__[k] = us.states.lookup(v)
            else:
                self.__dict__[k] = v

    @property
    def is_federal(self):
        return self.jurisdiction == "federal"

    @property
    def office_verbose(self):
        return EXECUTIVE_OFFICE_NAMES.get(self.office, "Unknown office")

    def __repr__(self):
        return "<ExecutiveSeat: {}>".format(self.__str__())

    def __str__(self):
        if self.is_federal:
            return "U.S. {}".format(self.office_verbose)
        else:
            return "{} {}".format(self.state.name, self.office_verbose)


class HeadOfGovernmentSeat(ExecutiveSeat):
    head_of_government = True



# Imports from other dependencies.
import us


class GeneralElection(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if k == "state":
                self.__dict__[k] = us.states.lookup(v)
            else:
                self.__dict__[k] = v

    def __repr__(self):
        return "<GeneralElection: {}>".format(self.__str__())

    def __str__(self):
        return "{} ({:%b. %d, %Y})".format(self.state.name, self.election_date)



# Imports from us-elections.
from elections.models.seats.executive import ExecutiveSeat
from elections.models.seats.executive import HeadOfGovernmentSeat
from elections.models.seats.house import HouseSeat
from elections.models.seats.senate import SenateSeat


__all__ = ["ExecutiveSeat", "HeadOfGovernmentSeat", "HouseSeat", "SenateSeat"]



raw_alternative_areas = [
    {
        "is_continental": False,
        "name": "Democrats Abroad",
        "is_obsolete": False,
        "capital": None,
        "ap_abbr": "Democrats Abroad",
        "time_zones": None,
        "fips": None,
        "abbr": "Democrats Abroad",
        "statehood_year": None,
        "is_contiguous": False,
        "capital_tz": None,
        "is_territory": False,
        "slug": "democrats-abroad",
    }
]


class ElectoralArea(object):
    repr_label = "ElectoralArea"

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            self.__dict__[k] = v

        try:
            import jellyfish

            self.__dict__["name_metaphone"] = jellyfish.metaphone(
                self.__dict__["name"]
            )
        except:
            pass

    def __repr__(self):
        return "<{}: {}>".format(self.repr_label, self.__str__())

    def __str__(self):
        return self.name


alternative_areas = dict(
    (area.get("slug").lower(), ElectoralArea(**area))
    for area in raw_alternative_areas
)



# The first Presidential election was held in 1788, and such elections
# have been held every four years since.
FIRST_PRESIDENTIAL_ELECTION = 1788
PRESIDENTIAL_ELECTION_INTERVAL = 4


# Years at the "top" of a cycle are "presidential years".
# Years at the "bottom" of a cycle are "midterm years."
# We have placeholders to describe the other (odd-numbered) years, but
# we could stand to have better descriptors for these.
PRESIDENTIAL_YEAR_TYPE = "presidential"
ODD_YEAR_ONE_TYPE = "odd-year-1"
MIDTERM_YEAR_TYPE = "midterm"
ODD_YEAR_TWO_TYPE = "odd-year-3"

ELECTION_YEAR_TYPES = [
    PRESIDENTIAL_YEAR_TYPE,
    ODD_YEAR_ONE_TYPE,
    MIDTERM_YEAR_TYPE,
    ODD_YEAR_TWO_TYPE,
]



# Imports from us-elections.
from elections.utils.alternative_areas import alternative_areas


# Imports from other dependencies.
import us


class PrimaryElection(object):
    is_runoff = False
    repr_label = "PrimaryElection"

    def __init__(self, **kwargs):
        if self.is_runoff:
            date_field = "runoff_election_date"
        else:
            date_field = "election_date"

        for k, v in kwargs.items():
            if k == "state":
                state_object = us.states.lookup(v)
                self.__dict__[k] = (
                    state_object
                    if state_object is not None
                    else alternative_areas[v.lower()]
                )
            elif k == "election_notes":
                self.__dict__[k] = v
            elif k == date_field:
                self.__dict__["election_date"] = v
            elif k != date_field and k.endswith("election_date"):
                pass
            else:
                self.__dict__[k] = v

    def __repr__(self):
        return "<{}: {}>".format(self.repr_label, self.__str__())

    def __str__(self):
        return "{} ({:%b. %d, %Y})".format(self.state.name, self.election_date)


class DemocraticPrimaryElection(PrimaryElection):
    repr_label = "DemocraticPrimaryElection"


class RepublicanPrimaryElection(PrimaryElection):
    repr_label = "RepublicanPrimaryElection"



# Imports from other dependencies.
import us


class SenateSeat(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if k == "state":
                self.__dict__[k] = us.states.lookup(v)
            else:
                self.__dict__[k] = v

    @property
    def is_federal(self):
        return self.jurisdiction == "federal"

    def __repr__(self):
        return "<SenateSeat: {}>".format(self.__str__())

    def __str__(self):
        if self.is_federal:
            return "{} U.S. Senate seat, class {}".format(
                self.state.name, self.senate_class
            )
        else:
            return "{} state senate seat, class {}".format(
                self.state.name, self.senate_class
            )



# Imports from other dependencies.
import us


# Imports from us-elections.
from elections.utils.formatters import ordinalize


class HouseSeat(object):
    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if k == "state":
                self.__dict__[k] = us.states.lookup(v)
            else:
                self.__dict__[k] = v

    @property
    def is_federal(self):
        return self.jurisdiction == "federal"

    @property
    def is_at_large(self):
        return not self.district

    @property
    def district_name(self):
        if self.is_at_large:
            return "at-large district"
        return "{} district".format(ordinalize(int(self.district)))

    def __repr__(self):
        return "<HouseSeat: {}>".format(self.__str__())

    def __str__(self):
        if self.is_federal:
            return "{} U.S. House seat, {}".format(
                self.state.name, self.district_name
            )
        else:
            return "{} state house seat, {}".format(
                self.state.name, self.district_name
            )



# Imports from us-elections.
from elections.models.elections import DemocraticPrimaryElection
from elections.models.elections.primary_runoff import (
    DemocraticPrimaryRunoffElection,
)
from elections.models.elections import GeneralElection
from elections.models.elections import PrimaryElection
from elections.models.elections.primary_runoff import PrimaryRunoffElection
from elections.models.elections import RepublicanPrimaryElection
from elections.models.elections.primary_runoff import (
    RepublicanPrimaryRunoffElection,
)
from elections.models.electoral_votes import DistrictElectoralZone
from elections.models.electoral_votes import StateElectoralZone
from elections.models.seats import HeadOfGovernmentSeat
from elections.models.seats import HouseSeat
from elections.models.seats import SenateSeat


PARTISAN_ELECTION_TYPES = {
    "democratic": {
        "primary": DemocraticPrimaryElection,
        "primary_runoff": DemocraticPrimaryRunoffElection,
    },
    "republican": {
        "primary": RepublicanPrimaryElection,
        "primary_runoff": RepublicanPrimaryRunoffElection,
    },
}

TO_WORD_ORDINALS = ["first", "second", "third"]


class ChamberFilterableList(list):
    @property
    def senate(self):
        return [
            seat for seat in self.__iter__() if isinstance(seat, SenateSeat)
        ]

    @property
    def house(self):
        return [
            seat for seat in self.__iter__() if isinstance(seat, HouseSeat)
        ]


class ExecutiveSeatFilterableList(list):
    @property
    def chief(self):
        chiefs = [
            seat
            for seat in self.__iter__()
            if isinstance(seat, HeadOfGovernmentSeat)
        ]
        if len(chiefs) > 1:
            raise ValueError(
                "State and federal executive branches have only one "
                "Head of Government."
            )
        return chiefs[0] if len(chiefs) == 1 else None

    @property
    def lower_seats(self):
        return [
            seat
            for seat in self.__iter__()
            if not isinstance(seat, HeadOfGovernmentSeat)
        ]


class StateFilterableList(list):
    def add_state(self, state_name, state_government_obj):
        setattr(self, state_name, state_government_obj)
        self.append(state_government_obj)


class PartisanElectionFilterableList(list):
    @property
    def democratic(self):
        return [
            election
            for election in self.__iter__()
            if any(
                [
                    isinstance(election, Klass)
                    for election_type, Klass in PARTISAN_ELECTION_TYPES[
                        "democratic"
                    ].items()
                    if Klass
                ]
            )
        ]

    @property
    def republican(self):
        return [
            election
            for election in self.__iter__()
            if any(
                [
                    isinstance(election, Klass)
                    for election_type, Klass in PARTISAN_ELECTION_TYPES[
                        "republican"
                    ].items()
                    if Klass
                ]
            )
        ]


class ElectionTypeFilterableList(list):
    @property
    def general_elections(self):
        return [
            election
            for election in self.__iter__()
            if isinstance(election, GeneralElection)
        ]

    @property
    def primaries(self):
        return PartisanElectionFilterableList(
            [
                election
                for election in self.__iter__()
                if isinstance(election, PrimaryElection)
                and not isinstance(election, PrimaryRunoffElection)
            ]
        )

    @property
    def primary_runoffs(self):
        return PartisanElectionFilterableList(
            [
                election
                for election in self.__iter__()
                if isinstance(election, PrimaryRunoffElection)
            ]
        )

    @property
    def presidential_primaries(self):
        return PartisanElectionFilterableList(
            [
                election
                for election in self.__iter__()
                if election.election_date.year % 4 == 0
                and isinstance(election, PrimaryElection)
                and not isinstance(election, PrimaryRunoffElection)
                and getattr(election, "election_level", "")
                in ["president", "all"]
            ]
        )

    @property
    def downticket_primaries(self):
        return PartisanElectionFilterableList(
            [
                election
                for election in self.__iter__()
                if isinstance(election, PrimaryElection)
                and not isinstance(election, PrimaryRunoffElection)
                and getattr(election, "election_level", "all")
                in ["all", "downticket"]
            ]
        )


class ElectoralVoteFilterableList(list):
    def add_zone(self, electoral_zone):
        state_slug = electoral_zone.state.name.replace(" ", "_").lower()

        if isinstance(electoral_zone, StateElectoralZone):
            setattr(self, state_slug, electoral_zone)
        else:
            setattr(
                self,
                "{}_{}_district".format(
                    state_slug,
                    TO_WORD_ORDINALS[int(electoral_zone.district) - 1],
                ),
                electoral_zone,
            )
        self.append(electoral_zone)

    @property
    def statewide(self):
        return [
            electoral_zone
            for electoral_zone in self.__iter__()
            if isinstance(electoral_zone, StateElectoralZone)
        ]

    @property
    def by_district(self):
        return [
            electoral_zone
            for electoral_zone in self.__iter__()
            if isinstance(electoral_zone, DistrictElectoralZone)
        ]



# Imports from other dependencies.
import us


# Imports from us-elections.
from elections.models.governments.government_branch import ExecutiveBranch
from elections.models.governments.government_branch import GovernmentBranch
from elections.models.governments.government_branch import LegislativeBranch
from elections.models.governments.constants import BRANCHES
from elections.models.governments.constants import LEGISLATIVE
from elections.models.governments.constants import EXECUTIVE


class GovernmentLevel(object):
    jurisdiction = None

    verbose_class_name = "Government"

    def __init__(self, **kwargs):
        for k, v in kwargs.items():
            if k in BRANCHES:
                if k == EXECUTIVE:
                    branch_klass = ExecutiveBranch
                else:
                    branch_klass = GovernmentBranch

                self.__dict__[k] = branch_klass(
                    branch_type=k, jurisdiction=self.jurisdiction, seats=v
                )

            else:
                self.__dict__[k] = v

    def get_verbose_class_name(self):
        return self.verbose_class_name

    def __repr__(self):
        return "<{}: {}>".format(self.get_verbose_class_name(), self.__str__())

    def __str__(self):
        return getattr(self.jurisdiction, "name", "Undefined")

    @property
    def abbreviation(self):
        return self.jurisdiction.__dict__.get("ap_abbr", "??")


class StateGovernment(GovernmentLevel):
    verbose_class_name = "StateGovernment"


class FederalGovernment(GovernmentLevel):
    verbose_class_name = "FederalGovernment"

    def __init__(self, **kwargs):
        government_jurisdiction = kwargs.get("jurisdiction", self.jurisdiction)
        congress_seats = kwargs.pop(LEGISLATIVE, [])
        if congress_seats:
            self.__dict__[LEGISLATIVE] = LegislativeBranch(
                branch_type=LEGISLATIVE,
                jurisdiction=government_jurisdiction,
                seats=congress_seats,
                body_name="Congress",
            )

        raw_electoral_votes = kwargs.pop("electoral_votes", None)

        super(FederalGovernment, self).__init__(**kwargs)

        if self.legislative:
            self.congress = self.legislative

        if self.executive and not self.executive.is_empty:
            if raw_electoral_votes:
                self.executive.chief.electoral_votes = raw_electoral_votes

            self.president = self.executive.chief
        elif self.executive and self.executive.is_empty:
            self.president = None



# Imports from other dependencies.
import us


# Imports from us-elections.
from elections.utils.formatters import ordinalize


class ElectoralZone(object):
    def __init__(self, **kwargs):
        self.is_federal = True

        for k, v in kwargs.items():
            if k == "state":
                self.__dict__[k] = us.states.lookup(v)
            elif k == "district" and v:
                self.__dict__[k] = v
            elif k == "elector_count":
                self.__dict__[k] = int(v)
            else:
                self.__dict__[k] = v

    @property
    def is_statewide(self):
        return not self.district

    @property
    def district_name(self):
        if self.is_statewide:
            return "statewide vote"
        return "{} district vote".format(ordinalize(int(self.district)))

    def __repr__(self):
        return "<ElectoralZone: {}>".format(self.__str__())

    def __str__(self):
        return "{} {}".format(self.state.name, self.district_name)


class DistrictElectoralZone(ElectoralZone):
    pass


class StateElectoralZone(ElectoralZone):
    pass



def get_party(party_code, parties):
    if not party_code:
        return None
    matched = list(
        filter(lambda d: d.ap_code.lower() == party_code.lower(), parties)
    )
    if len(matched) > 0:
        return matched[0]
    return None



# Imports from python.
import pickle


# Imports from other dependencies.
import us


# Imports from us-elections.
from elections.exceptions import InvalidYear
from elections.models.elections import GeneralElection
from elections.models.elections import DemocraticPrimaryElection
from elections.models.elections import DemocraticPrimaryRunoffElection
from elections.models.elections import PrimaryElection
from elections.models.elections import PrimaryRunoffElection
from elections.models.elections import RepublicanPrimaryElection
from elections.models.elections import RepublicanPrimaryRunoffElection
from elections.models.electoral_votes import DistrictElectoralZone
from elections.models.electoral_votes import StateElectoralZone
from elections.models.governments.constants import LEGISLATIVE
from elections.models.governments.constants import EXECUTIVE
from elections.models.governments.constants import JUDICIAL
from elections.models.governments import FederalGovernment
from elections.models.governments import StateGovernment
from elections.models.seats import ExecutiveSeat
from elections.models.seats import HeadOfGovernmentSeat
from elections.models.seats import HouseSeat
from elections.models.seats import SenateSeat
from elections.models.parties import Party
from elections.utils.alternative_areas import alternative_areas
from elections.utils.getters import get_party
from elections.utils.builts import ChamberFilterableList
from elections.utils.builts import ElectionTypeFilterableList
from elections.utils.builts import ElectoralVoteFilterableList
from elections.utils.builts import ExecutiveSeatFilterableList
from elections.utils.builts import StateFilterableList
from elections.utils.constants import ELECTION_YEAR_TYPES
from elections.utils.constants import FIRST_PRESIDENTIAL_ELECTION
from elections.utils.constants import MIDTERM_YEAR_TYPE
from elections.utils.constants import PRESIDENTIAL_ELECTION_INTERVAL
from elections.utils.constants import PRESIDENTIAL_YEAR_TYPE
from elections.utils.data_loaders import load_pickled_data


def load_years():
    years = []
    with load_pickled_data("data/years.pkl") as pklfile:
        for d in pickle.load(pklfile):
            years.append(d)
    return years


CHAMBER_ORDER = ["senate", "house"]


class ElectionYear(object):
    def __init__(self, year, **kwargs):
        year = str(year)
        year_as_int = int(year)
        if year not in load_years():
            raise InvalidYear(
                "No data available for the {} election cycle".format(year)
            )
        self.year = year

        self.include_states = True

        for k, v in kwargs.items():
            if k == "include_states":
                self.include_states = bool(v)
            else:
                self.__dict__[k] = v

        self.incumbent_parties = self.load_parties()

        self.election_year_type = ELECTION_YEAR_TYPES[
            (year_as_int - FIRST_PRESIDENTIAL_ELECTION)
            % PRESIDENTIAL_ELECTION_INTERVAL
        ]

        patched_federal_jurisdiction = us.unitedstatesofamerica
        patched_federal_jurisdiction.ap_abbr = "U.S."

        federal_government_params = {
            "jurisdiction": patched_federal_jurisdiction,
            LEGISLATIVE: (
                self.load_federal_legislative_seats()
                if self.election_year_type
                in [PRESIDENTIAL_YEAR_TYPE, MIDTERM_YEAR_TYPE]
                else []
            ),
            EXECUTIVE: (
                self.load_federal_executive_seats()
                if self.election_year_type == PRESIDENTIAL_YEAR_TYPE
                else []
            ),
            JUDICIAL: [],  # No federal judiciary members are elected.
        }

        if self.election_year_type == PRESIDENTIAL_YEAR_TYPE:
            federal_government_params[
                "electoral_votes"
            ] = self.load_electoral_votes()

        self.federal = FederalGovernment(**federal_government_params)

        if self.include_states:
            self.states = self.load_states()

        self.elections = self.load_elections()

    def load_parties(self):
        parties = []
        with load_pickled_data("data/parties.pkl") as pklfile:
            for d in pickle.load(pklfile):
                parties.append(Party(**d))
        return parties

    def assign_parties_to_seats(self, seats):
        for seat in seats:
            holding_party = get_party(
                getattr(seat, "incumbent_party", None), self.incumbent_parties
            )
            if holding_party:
                holding_party.add_held_seat(seat)
                seat.incumbent_party = holding_party
        # filter out parties that don't hold any seats
        self.incumbent_parties = list(
            filter(lambda d: d.holds_seats, self.incumbent_parties)
        )
        return seats

    def load_federal_legislative_seats(self):
        seats = []
        with load_pickled_data(
            "data/{}/federal-legislative-seats.pkl".format(self.year)
        ) as pklfile:
            for d in pickle.load(pklfile):
                if d["seat_type"] == "house":
                    seats.append(HouseSeat(**d))
                elif d["seat_type"] == "senate":
                    seats.append(SenateSeat(**d))

        sorted_seats = sorted(
            seats,
            key=lambda x: int(getattr(x, "district", "0"))
            if hasattr(x, "district") and x.district is not None
            else 0,
        )
        sorted_seats = sorted(
            sorted_seats, key=lambda x: len(getattr(x, "senate_class", "IIII"))
        )
        sorted_seats = sorted(sorted_seats, key=lambda x: x.state.name)
        sorted_seats = sorted(
            sorted_seats, key=lambda x: CHAMBER_ORDER.index(x.seat_type)
        )

        return self.assign_parties_to_seats(
            ChamberFilterableList(sorted_seats)
        )

    def load_federal_executive_seats(self):
        seats = ExecutiveSeatFilterableList()
        with load_pickled_data(
            "data/{}/federal-executive-seats.pkl".format(self.year)
        ) as pklfile:
            for d in pickle.load(pklfile):
                if d["office"] == "president":
                    seats.append(HeadOfGovernmentSeat(**d))
                else:
                    seats.append(ExecutiveSeat(**d))

        seats = self.assign_parties_to_seats(seats)

        return seats

    def load_electoral_votes(self):
        if self.election_year_type != PRESIDENTIAL_YEAR_TYPE:
            raise ValueError(
                "Only presidential election years have electoral votes."
            )

        vote_zones = ElectoralVoteFilterableList()
        with load_pickled_data(
            "data/{}/electoral-votes.pkl".format(self.year)
        ) as pklfile:
            for d in pickle.load(pklfile):
                if d["type"] == "statewide":
                    vote_zones.add_zone(StateElectoralZone(**d))
                else:
                    vote_zones.add_zone(DistrictElectoralZone(**d))

        # vote_zones = self.assign_states_to_vote_zones(vote_zones)

        return vote_zones

    def load_states(self):
        state_list = StateFilterableList()

        for state in us.STATES:
            if state.statehood_year:  # Taxation w/o representation! Nix D.C.
                state.slug = state.name.replace(" ", "_").lower()
                state_list.add_state(
                    state.slug,
                    StateGovernment(
                        **{
                            "jurisdiction": state,
                            LEGISLATIVE: (
                                self.load_state_legislative_seats(
                                    state.abbr.lower()
                                )
                            ),
                            EXECUTIVE: (
                                self.load_state_executive_seats(
                                    state.abbr.lower()
                                )
                            ),
                            JUDICIAL: [],
                        }
                    ),
                )

        return state_list

    def load_state_legislative_seats(self, state_abbr):
        seats = []
        with load_pickled_data(
            "data/{}/{}-legislative-seats.pkl".format(self.year, state_abbr)
        ) as pklfile:
            for d in pickle.load(pklfile):
                if d["seat_type"] == "house":
                    seats.append(HouseSeat(**d))
                elif d["seat_type"] == "senate":
                    seats.append(SenateSeat(**d))

        sorted_seats = sorted(
            seats, key=lambda x: CHAMBER_ORDER.index(x.seat_type)
        )

        return self.assign_parties_to_seats(
            ChamberFilterableList(sorted_seats)
        )

    def load_state_executive_seats(self, state_abbr):
        seats = ExecutiveSeatFilterableList()
        with load_pickled_data(
            "data/{}/{}-executive-seats.pkl".format(self.year, state_abbr)
        ) as pklfile:
            for d in pickle.load(pklfile):
                if d["office"] == "governor":
                    seats.append(HeadOfGovernmentSeat(**d))
                else:
                    seats.append(ExecutiveSeat(**d))

        return self.assign_parties_to_seats(seats)

    def load_elections(self):
        elections = []

        with load_pickled_data(
            "data/{}/elections.pkl".format(self.year)
        ) as pklfile:
            for d in pickle.load(pklfile, encoding="latin1"):
                if d["election_type"] == "general":
                    elections.append(GeneralElection(**d))
                elif d["election_type"] == "primary":
                    if d["election_party"] == "dem":
                        if d["election_date"] is not None:
                            elections.append(DemocraticPrimaryElection(**d))
                        if d["runoff_election_date"] is not None:
                            elections.append(
                                DemocraticPrimaryRunoffElection(**d)
                            )
                    elif d["election_party"] == "gop":
                        if d["election_date"] is not None:
                            elections.append(RepublicanPrimaryElection(**d))
                        if d["runoff_election_date"] is not None:
                            elections.append(
                                RepublicanPrimaryRunoffElection(**d)
                            )
                    else:
                        if d["election_date"] is not None:
                            elections.append(PrimaryElection(**d))
                        if d["runoff_election_date"] is not None:
                            elections.append(PrimaryRunoffElection(**d))

        sorted_elections = sorted(elections, key=lambda x: x.state.name)
        sorted_elections = sorted(
            sorted_elections, key=lambda x: x.election_date
        )

        return ElectionTypeFilterableList(sorted_elections)

    def elections_for_state(self, state_name_raw):
        state = us.states.lookup(state_name_raw)

        if not state:
            state = alternative_areas.get(state_name_raw, None)

        if not state:
            raise ValueError(f'No state found for query "{state_name_raw}".')

        return ElectionTypeFilterableList(
            [
                election
                for election in self.elections
                if election.state == state
            ]
        )

    def __repr__(self):
        return "<ElectionYear: {}>".format(self.__str__())

    def __str__(self):
        if self.election_year_type in [
            PRESIDENTIAL_YEAR_TYPE,
            MIDTERM_YEAR_TYPE,
        ]:
            return "{} ({} cycle)".format(
                self.year, self.election_year_type.capitalize()
            )
        return self.year



# Imports from us-elections.
from elections.models.elections.general import GeneralElection
from elections.models.elections.primary import DemocraticPrimaryElection
from elections.models.elections.primary import PrimaryElection
from elections.models.elections.primary import RepublicanPrimaryElection
from elections.models.elections.primary_runoff import (
    DemocraticPrimaryRunoffElection,
)
from elections.models.elections.primary_runoff import PrimaryRunoffElection
from elections.models.elections.primary_runoff import (
    RepublicanPrimaryRunoffElection,
)


__all__ = [
    "DemocraticPrimaryElection",
    "DemocraticPrimaryRunoffElection",
    "GeneralElection",
    "PrimaryElection",
    "PrimaryRunoffElection",
    "RepublicanPrimaryElection",
    "RepublicanPrimaryRunoffElection",
]
